package dstm2.util;

import dstm2.AtomicArray;
import java.io.IOException;
import java.util.AbstractSet;
import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;

import dstm2.Thread;
import dstm2.atomic;
import dstm2.factory.Factory;

public class HashMap2<V extends dstm2.AtomicSuperClass> {

    /**
     * The default initial capacity - MUST be a power of two.
     */
    static final int DEFAULT_INITIAL_CAPACITY = 4096;
    /**
     * The maximum capacity, used if a higher value is implicitly specified
     * by either of the constructors with arguments.
     * MUST be a power of two <= 1<<30.
     */
    static final int MAXIMUM_CAPACITY = 1 << 30;
    /**
     * The load factor used when none specified in constructor.
     **/
    static final float DEFAULT_LOAD_FACTOR = 0.75f;
 
    tableHolder tableholder;
    final private Factory<tableHolder> factoryTable;
    final private Factory<TEntry> factoryTEntry;
    final private Factory<Wrapper> factorywrapper;
    
    int capacity;
    int threshold;
    final float loadFactor;

    public HashMap2(int initialCapacity, float loadFactor) {
        factoryTEntry = Thread.makeFactory(TEntry.class);
        factorywrapper = Thread.makeFactory(Wrapper.class);
        factoryTable = Thread.makeFactory(tableHolder.class);
        if (initialCapacity < 0) {
            throw new IllegalArgumentException("Illegal initial capacity: " +
                    initialCapacity);
        }
        if (initialCapacity > MAXIMUM_CAPACITY) {
            initialCapacity = MAXIMUM_CAPACITY;
        }
        if (loadFactor <= 0 || Float.isNaN(loadFactor)) {
            throw new IllegalArgumentException("Illegal load factor: " +
                    loadFactor);

        // Find a power of 2 >= initialCapacity
        }
        int tmpcapacity = 1;
        while (tmpcapacity < initialCapacity) {
            tmpcapacity <<= 1;
        }
        capacity = tmpcapacity;
        this.loadFactor = loadFactor;
        threshold = (int) (capacity * loadFactor);
        //threshold = (int) (capacity * loadFactor);
        tableholder = factoryTable.create();
        AtomicArray<Wrapper> table = new dstm2.AtomicArray<Wrapper>(Wrapper.class, capacity);
        tableholder.setTable(table);
        for (int i = 0; i < capacity; i++) {
            tableholder.getTable().set(i, factorywrapper.create());
            tableholder.getTable().get(i).setTEntry(factoryTEntry.create());
        }
        init();
    }

  
    public HashMap2(int initialCapacity) {
        this(initialCapacity, DEFAULT_LOAD_FACTOR);
    }

    /**
     * Constructs an empty <tt>HashMap</tt> with the default initial capacity
     * (16) and the default load factor (0.75).
     */
    public HashMap2() {
        this(DEFAULT_INITIAL_CAPACITY, DEFAULT_LOAD_FACTOR);
    }

 
    void init() {
    }

    public dstm2.AtomicArray<Wrapper>/*dstm2.AtomicArray<TEntry>*/ getBuckets() {
        return tableholder.getTable();
    }
    /**
     * Value representing null keys inside tables.
     */
    static final Object NULL_KEY = new Object();

    /**
     * Returns internal representation for key. Use NULL_KEY if key is null.
     */
    static <T> T maskNull(T key) {
        return key == null ? (T) NULL_KEY : key;
    }

    /**
     * Returns key represented by specified internal representation.
     */
    static <T> T unmaskNull(T key) {
        return (key == NULL_KEY ? null : key);
    }
   
    private static final boolean useNewHash;
    

    static {
        useNewHash = false;
    }

    private static int oldHash(int h) {
        h += ~(h << 9);
        h ^= (h >>> 14);
        h += (h << 4);
        h ^= (h >>> 10);
        return h;
    }

    private static int newHash(int h) {
        h ^= (h >>> 20) ^ (h >>> 12);
        return h ^ (h >>> 7) ^ (h >>> 4);
    }


    public static int hash(int h) {
        return useNewHash ? newHash(h) : oldHash(h);
    }

    static int hash(Object key) {
        return hash(key.hashCode());
    }

    
    static boolean eq(Object x, Object y) {
        return x == y || x.equals(y);
    }

  
    public static int indexFor(int h, int length) {
        return h & (length - 1);
    }

 
 
    public V get(int key) {

        int hash = hash(key);
        TEntry<V> e = tableholder.getTable().get(indexFor(hash, capacity)).getTEntry();

        if (tableholder.getTable().get(indexFor(hash, capacity)).getTEntry().getHash() == hash) {
            return (V)tableholder.getTable().get(indexFor(hash, capacity)).getTEntry().getValue();
        }
        return null;
    }


    public boolean containsKey(int key) {
        int hash = hash(key);
        int i = indexFor(hash, capacity);
        if (tableholder.getTable().get(i).getTEntry().getHash() == hash) {
            return true;
        }
        return false;
    }

   
    TEntry<V> getEntry(int key) {
        int hash = hash(key);
        int i = indexFor(hash, capacity);
        TEntry<V> e = tableholder.getTable().get(i).getTEntry();
        if (e.getHash() == hash) {
            return e;
        }
        return null;
    }

    public V put(int key, V value) {
        int hash = hash(key);
        int i = indexFor(hash, capacity);
        TEntry<V> e = tableholder.getTable().get(i).getTEntry();
        if (tableholder.getTable().get(i).getTEntry().getHash() == hash) {
            V oldValue = (V) tableholder.getTable().get(i).getTEntry().getValue();
            tableholder.getTable().get(i).getTEntry().setValue(value);
            return oldValue;
        }
        addEntry(hash, value, i);
        return null;

    }
    
    void resize(int newCapacity) {
        int oldCapacity = capacity;
        if (oldCapacity == MAXIMUM_CAPACITY) {
            threshold = Integer.MAX_VALUE;
            return;

        }
        dstm2.AtomicArray<Wrapper> newTable = new dstm2.AtomicArray<Wrapper>(Wrapper.class, newCapacity);
        for (int i = 0;
                i < newCapacity;
                i++) {
            if (i < capacity) {
                newTable.set(i, tableholder.getTable().get(i)/*factorywrapper.create()*/);
                newTable.get(i).setTEntry(factoryTEntry.create());
            } else {
                newTable.set(i, factorywrapper.create());
                newTable.get(i).setTEntry(factoryTEntry.create());
            }
        }
        transfer(newTable, newCapacity);
        tableholder.setTable(newTable);
        threshold = (int) (newCapacity * loadFactor);

        capacity = newCapacity;
    }


    void transfer(dstm2.AtomicArray<Wrapper> newTable, int nc) {
        int newCapacity = nc;
        for (int j = 0; j < capacity; j++) {
            TEntry<V> e = tableholder.getTable().get(j).getTEntry();
            if (e != null) {
                tableholder.getTable().set(j, null);
                int i = indexFor(e.getHash(), newCapacity);
                newTable.get(i).setTEntry(e);
            }

        }
    }


    public V remove(int key) {
        V e = removeEntryForKey(key);
        return (e == null ? null : e);
    }

    V removeEntryForKey(int key) {
        int hash = hash(key);
        int i = indexFor(hash, capacity);
        //TEntry<V> prev = table.get(i);
        V e = (V) tableholder.getTable().get(i).getTEntry().getValue();
        if (tableholder.getTable().get(i).getTEntry().getHash() == hash) {
            tableholder.getTable().get(i).setTEntry(null);
        }
        return e;
    }

    public void clear() {
        for (int i = 0; i<capacity; i++) {
            tableholder.getTable().set(i, null);
        }

    }

    public boolean containsValue(Object value) {
        if (value == null) {
            return containsNullValue();
        }
        for (int i = 0; i <
                capacity; i++) {
            TEntry e = tableholder.getTable().get(i).getTEntry();
            if (value.equals(e.getValue())) {
                return true;
            }

        }

        return false;
    }

    /**
     * Special-case code for containsValue with null argument
     **/
    private boolean containsNullValue() {
        for (int i = 0; i < capacity; i++) {
            TEntry e = tableholder.getTable().get(i).getTEntry();
            if (e.getValue() == null) {
                return true;
            }
        }
        return false;
    }

  
 
    void addEntry(int hash, V value, int bucketIndex) {
        tableholder.getTable().get(bucketIndex).getTEntry().setHash(hash);
        tableholder.getTable().get(bucketIndex).getTEntry().setValue(value);
    }


    void createEntry(int hash, V value, int bucketIndex) {
        //TEntry<V> e = table.get(bucketIndex);
        TEntry<V> e = tableholder.getTable().get(bucketIndex).getTEntry();
        TEntry<V> n = factoryTEntry.create();
        n.setHash(hash);
        n.setValue(value);
        tableholder.getTable().get(bucketIndex).setTEntry(n);
    }
    
      @atomic
    public interface transactionalintfield {

        int getValue();

        void setValue(int val);
    }

    @atomic
    public interface tableHolder {

        dstm2.AtomicArray<Wrapper> getTable();

        void setTable(dstm2.AtomicArray<Wrapper> table);
    }

    @atomic
    public interface Wrapper {

        TEntry getTEntry();

        void setTEntry(TEntry n);
    }

    @atomic
    public interface TEntry<V> {

        int getHash();

        V getValue();

        void setHash(int h);

        void setValue(V v);
    }

}
