/*
 * Copyright (c) 2007, Solido Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name of Solido Systems nor the names of its contributors may be
 * used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package tuplesoup.test;

import tuplesoup.core.*;
import tuplesoup.filter.*;

import dstm2.Defaults;
import dstm2.Init;
import java.util.*;
import dstm2.Thread;
import java.io.*;

public class ParallelPerformanceTest extends BasicTest implements Runnable {

    long writetime;
    long readtime;
    long randomtime;

    public ParallelPerformanceTest(int records, int threadnum, String path) {
        try {
            for (int i = 1; i < 2; i++) {
                outbr("Running Parallel DualFileTable Performance test");
                outbr(1, i + " x " + (records / i) + " Large records");
                Table table;

                table = new DualFileTable("Performance-test", path, Table.PAGED);
		benchmarkSingleThreadedLargeWrite(table, records);
                outbr("Table Ready");
                benchmark(table, threadnum, records);
                table.close();
                table.deleteFiles();

            }


        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public static void main(String[] args) {
	 int argc = 0;
	 int records = 500;
	 int numThreads = Defaults.THREADS;
	 boolean inevitable = Defaults.INEVITABLE;
	 String managerClassName = Defaults.MANAGER;
	 String adapterClassName = Defaults.ADAPTER;
	 String basedir = "/scratch/TransactionalIO/tuplesoupfilesv0/";
	 try {
      	        while (argc < args.length) {
		        String option = args[argc++];
	        	if (option.equals("-m")){
        			 managerClassName = args[argc];
		        }
        		else if (option.equals("-records")){
				records = Integer.parseInt(args[argc]);
			}
        		else if (option.equals("-inevitable")){
	          		inevitable = Boolean.parseBoolean(args[argc]);
	        	}
	        	else if (option.equals("-t"))
		                numThreads = Integer.parseInt(args[argc]);
	        	else if (option.equals("-a"))
		                adapterClassName = args[argc];
	        	else if (option.equals("-basedir"))
		                basedir = args[argc];
	        	else
		          reportUsageErrorAndDie();
	                argc++;
 	      }
	 } catch (NumberFormatException e) {
	      System.out.println("Expected a number: " + args[argc]);
	      System.exit(0);
	 } 
         Init.init(managerClassName, inevitable,adapterClassName);
	 System.out.println("Numberof Threads " + numThreads);
	
         new ParallelPerformanceTest(records, numThreads, basedir);
    }

    public void benchmark(Table table, int threadcount, int records) throws Exception {
        writetime = 0;
        readtime = 0;
        randomtime = 0;
        List<Thread> lst = new ArrayList<Thread>();
        long starttime = System.currentTimeMillis();
        for (int i = 0; i < threadcount; i++) {
            Thread thr = new Thread(new ParallelThread(this, table, i + "", records));
            thr.start();
            lst.add(thr);
        }
        for (int i = 0; i < threadcount; i++) {
            lst.get(i).join();
        }
        long stoptime = System.currentTimeMillis();
        System.out.println("\n\nExectuin time: " + (long) (stoptime - starttime));
        long committed = Thread.totalCommitted;
        long total = Thread.totalTotal;
        if (total > 0) {
            System.out.printf("Committed: %d\nTotal: %d\nPercent committed: (%d%%)\n",
                    committed,
                    total,
                    (100 * committed) / total);
        } else {
            System.out.println("No transactions executed!");
        }

    }

    public void run() {
    }


    public void benchmarkSingleThreadedLargeWrite(Table table, int records) throws IOException {
        for (int i = 0; i < records; i++) {
            Row row = new Row(String.valueOf(i));
            row.put("key1", "foobarbazinevitable2");
            table.addRow(row);
        }
    }
        
    
    public void benchmarkAllTogether(Table table, int loops) throws IOException {
	for (int i=0; i<10000; i++){
	    int number = (int)(Math.random()*(loops/*+loops/10*/));
            String s = String.valueOf(number);
            Row row = new Row(s);
            row.put("key1", "foobarbazinevitable2");
	    table.addOrUpdateRow(row);
	}	

    }	

    private static void reportUsageErrorAndDie() {
        System.out.println("usage: dstm2.Main -b <benchmarkclass> [-m <managerclass>] [-t <#threads>] [-n <#time-in-ms>] [-e <experiment#>] [-a <adapter>]");
        System.exit(0);
    }

}
